package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.util.WidgetStateSignalManager;
import dev.pfaff.altargui.widgets.state.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.Signal;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.util.function.Supplier;

/**
 * A widget that replaces its child using the provided {@link #builder} whenever notified by signal.
 */
public record SignalBuilderWidget(Signal signal, Supplier<Widget> builder) implements Widget {
	@Override
	public @NotNull WidgetInstance<SignalBuilderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<SignalBuilderWidget> {
		private static final WidgetStateSignalManager<State, Signal> SIGNAL_MANAGER = new WidgetStateSignalManager<>(
				MethodHandles.lookup(),
				"signal", Signal.class,
				"listener",
				(state, signal) -> signal.listenAndInvoke(state::markNeedsRebuild)
		);

		private @Nullable Signal signal;
		private Supplier<Widget> builder;

		private @Nullable ListenerDisposer listener;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected @NotNull Widget build() {
			return this.builder.get();
		}

		@Override
		public void configure(StateContext stateContext, SignalBuilderWidget widget, boolean init) {
			if (this.builder != widget.builder) {
				this.builder = widget.builder;
				markNeedsRebuild();
			}
			SIGNAL_MANAGER.configure(this, widget.signal);
		}

		@Override
		protected void onMount() {
			super.onMount();

			SIGNAL_MANAGER.onMount(this);
		}

		@Override
		protected void onUnmount() {
			SIGNAL_MANAGER.onUnmount(this);
			super.onUnmount();
		}
	}
}
